/*
	BromScript - On Runtime Scripting Language.
    Copyright (C) 2012 - 2013  Alex Brouwer (Bromvlieg: http://brom.4o3.nl/)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Function.h"

using namespace Scratch;

namespace BromScript {
	extern int _fcount = 0;

	int Function::GetCount(){
		return _fcount;
	}

	Function::Function(Instance* bs){
		_fcount++;

		this->CurrentSourceFileLine = -1;
		this->Code = null;
		this->ForceReturn = false;
		this->IsCpp = false;
		this->Parent = null;
		this->BromScript = bs;
		this->CppFunc = null;
		this->Locals = new Table(bs);
		this->CurrentThisObject = null;
	}

	Function::~Function(){
		_fcount--;

		if (this->Code != null)
			delete this->Code;

		delete this->Locals;
	}

	void Function::SetVar(CString key, Variable* var, bool global){
		if (global){
			BromScript::Function* func = this;
			while(true){
				if (func == null){
					this->BromScript->Globals->Set(key, var);
					return;
				}else if (func->Locals->HasKey(key)){
					func->Locals->Set(key, var);
					return;
				}

				func = func->Parent;
			}
		}

		this->Locals->Set(key, var);
	}


	Variable* Function::GetVar(CString key){
		if (key == "this" && this->CurrentThisObject != null)
			return this->CurrentThisObject;

		BromScript::Function* func = this;
		while(true){
			if (func == null){
				return this->BromScript->Globals->Get(key);
			}else if (func->Locals->HasKey(key)){
				return func->Locals->Get(key);
			}

			func = func->Parent;
		}
	}

	Variable* Function::Run(){
		ArgumentData* args = new ArgumentData();
		args->BromScript = this->BromScript;
		args->Caller = this;
		args->Count = 0;

		Variable* ret = this->Run(args);
		delete args;

		return ret;
	}
	
	Variable* Function::Run(Variable* thisobject){
		Variable* oldthis = this->CurrentThisObject;
		this->CurrentThisObject = thisobject;

		ArgumentData* args = new ArgumentData();
		args->BromScript = this->BromScript;
		args->Caller = this;

		if (this->IsCpp) args->InsertVariable(0, thisobject);
		
		Variable* ret = this->Run(args);
		delete args;
		
		this->CurrentThisObject = oldthis;
		return ret;
	}

	Variable* Function::Run(List<Variable*>* vars){
		ArgumentData* args = new ArgumentData();
		args->BromScript = this->BromScript;
		args->Caller = this;
		args->SetVariableData(vars);

		Variable* ret = this->Run(args);
		delete args;

		return ret;
	}

	Variable* Function::Run(List<Variable*>* vars, Variable* thisobject){
		Variable* oldthis = this->CurrentThisObject;
		this->CurrentThisObject = thisobject;

		if (this->IsCpp) vars->Insert(0, thisobject);
		Variable* ret = this->Run(vars);

		this->CurrentThisObject = oldthis;
		return ret;
	}

	Variable* Function::Run(ArgumentData* args){
		if (this->BromScript->GetCurrentStackSize() >= 257){
			args->Error(CString("Stack error, cannot call '") + this->Name + "'");
			return this->BromScript->GC.RegisterVariable();
		}
		
		if (this->IsCpp){
			this->BromScript->EnterStack(this);
			Variable* ret = this->CppFunc(this->BromScript, args);
			this->BromScript->LeaveStack();

			if (ret == null) ret = this->BromScript->GC.RegisterVariable();
			else this->BromScript->GC.RegisterVariable(ret);
			
			return ret;
		}

		for (int i = 0; i < args->Count && i < this->Parameters.Count; i++){
			Variable* var = args->GetVariable(i);

			if (var->Type < Variable::VariableType::Table){
				Variable* paramvar = this->BromScript->GC.RegisterVariable();

				paramvar->Type = var->Type;
				paramvar->Value = var->Copy();
				paramvar->DeleteOnDestruct = var->DeleteOnDestruct;

				this->SetVar(this->Parameters[i], paramvar, false);
			}else
				this->SetVar(this->Parameters[i], var, false);
		}

		ExecuteData* data = new ExecuteData();
		data->BromScript = this->BromScript;
		data->Reader = new ByteReader();
		data->Reader->Func = this;
		data->Function = this;

		int oldline = this->CurrentSourceFileLine;
		this->CurrentSourceFileLine = -1;

		this->BromScript->EnterStack(this);
		Variable* ret = this->InternalRun(data);
		this->BromScript->LeaveStack();

		this->CurrentSourceFileLine = oldline;

		delete data;
		return ret;
	}
	
	Variable* Function::Run(ArgumentData* args, Variable* thisobject){
		Variable* oldthis = this->CurrentThisObject;
		this->CurrentThisObject = thisobject;

		if (this->IsCpp) args->InsertVariable(0, thisobject);
		Variable* ret = this->Run(args);

		this->CurrentThisObject = oldthis;
		return ret;
	}

	Variable* Function::InternalRun(ExecuteData* data){
		while(true){
			if (data->HasReturnValue || data->Breaking || data->Continueing || data->Function->ForceReturn)
				return data->Returning;

			if (data->BromScript->KillScriptThreaded)
				return data->BromScript->GC.RegisterVariable();

			Misc::ExecFuncs b = (Misc::ExecFuncs)data->Reader->ReadByte();

			switch(b){
			case Misc::ExecFuncs::GetVar: return Executer::GetVar(data); break;
			case Misc::ExecFuncs::GetTblIndex:
				if (data->FuncCallback) return Executer::GetTableIndex(data);
				else Executer::GetTableIndex(data);
				break; 

			case Misc::ExecFuncs::SetTblIndex:Executer::SetTableIndex(data); break; 

			case Misc::ExecFuncs::SetVar: Executer::SetVar(data); break;
			case Misc::ExecFuncs::LSetVar: Executer::LSetVar(data); break;

			case Misc::ExecFuncs::If: Executer::If(data); break;
			case Misc::ExecFuncs::ElseIf: Executer::If(data); break; // same as if
			case Misc::ExecFuncs::Else: Executer::Else(data); break;

			case Misc::ExecFuncs::For: Executer::For(data); break;
			case Misc::ExecFuncs::ForEach: Executer::ForEach(data); break;
			case Misc::ExecFuncs::While: Executer::While(data); break;

			case Misc::ExecFuncs::New: return Executer::New(data); break;
			case Misc::ExecFuncs::Function: Executer::CreateFunction(data); break;
			case Misc::ExecFuncs::AnonFunction: return Executer::CreateAnonFunction(data); break;

			case Misc::ExecFuncs::ExecuteFunction:
				if (data->FuncCallback) return Executer::ExecuteFunction(data);
				else Executer::ExecuteFunction(data);
				break; 

			case Misc::ExecFuncs::End:
				return data->BromScript->GC.RegisterVariable(); // lets give back null value.

			case Misc::ExecFuncs::MergeStart: return Executer::Merge(data);
			case Misc::ExecFuncs::Return: return Executer::Return(data);
			case Misc::ExecFuncs::Break: data->Breaking = true; return data->BromScript->GC.RegisterVariable();
			case Misc::ExecFuncs::Continue: data->Continueing = true; return data->BromScript->GC.RegisterVariable();

			case Misc::ExecFuncs::Bool: return Executer::GetBool(data);
			case Misc::ExecFuncs::String: return Executer::GetString(data);
			case Misc::ExecFuncs::Number: return Executer::GetNumber(data);
			case Misc::ExecFuncs::Table: return Executer::GetTable(data);
				
			case Misc::ExecFuncs::Skip: break;
			case Misc::ExecFuncs::CurrentLine:
				this->CurrentSourceFileLine = data->Reader->ReadInt();
				break;

			default:
				data->BromScript->Error(CString("Invalid bytecode detected! (") + CString((double)b) + ")");

				data->HasReturnValue = true;
				data->Returning = data->BromScript->GC.RegisterVariable();
				return data->Returning;
			}
		}
	}
}